from super_scad.d2.private.PrivateSquare import PrivateSquare
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2


class Rectangle(ScadWidget):
    """
    Widget for creating rectangles.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 size: Vector2 | None = None,
                 width: float | None = None,
                 depth: float | None = None,
                 center: bool = False):
        """
        Object constructor.

        :param size: The size of the rectangle.
        :param width: The width (the size along the x-axis) of the rectangle.
        :param depth: The depth (the size along the y-axis) of the rectangle.
        :param center: Whether the rectangle is centered at the origin.
        """
        ScadWidget.__init__(self, args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'size'}, {'width', 'depth'})
        admission.validate_required({'width', 'size'},
                                    {'depth', 'size'},
                                    {'center'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def size(self) -> Vector2:
        """
        Returns the size of the rectangle.
        """
        return Vector2(x=self.width, y=self.depth)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def width(self) -> float:
        """
        Returns the width (the size along the x-axis) of the rectangle.
        """
        if 'size' in self._args:
            return self.uc(self._args['size'].x)

        return self.uc(self._args['width'])

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def depth(self) -> float:
        """
        Returns the depth (the size along the y-axis) of the rectangle.
        """
        if 'size' in self._args:
            return self.uc(self._args['size'].y)

        return self.uc(self._args['depth'])

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the rectangle is centered at the origin.
        """
        return self._args['center']

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateSquare(size=self.size, center=self.center)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateTranslate import PrivateTranslate
from super_scad.type.Vector2 import Vector2


class Translate2D(ScadSingleChildParent):
    """
    Translates (moves) its child widget along the specified vector. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#translate.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 vector: Vector2 | None = None,
                 x: float | None = None,
                 y: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param vector: The vector over which the child widget is translated.
        :param x: The distance the child widget is translated to along the x-axis.
        :param y: The distance the child widget is translated to along the y-axis.
        :param child: The child widget to be translated.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'vector'}, {'x', 'y'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector2:
        """
        Returns the vector over which the child widget is translated.
        """
        return Vector2(self.x, self.y)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def x(self) -> float:
        """
        Returns distance the child widget is translated to along the x-axis.
        """
        if 'vector' in self._args:
            return self.uc(self._args['vector'].x)

        return self.uc(self._args.get('x', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def y(self) -> float:
        """
        Returns distance the child widget is translated to along the y-axis.
        """
        if 'vector' in self._args:
            return self.uc(self._args['vector'].y)

        return self.uc(self._args.get('y', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateTranslate(vector=Vector2(x=self.x, y=self.y), child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



import typing
from abc import ABC, abstractmethod
from typing import Any, Dict, List

from super_scad.scad import Length
from super_scad.scad.Context import Context
from super_scad.scad.Unit import Unit
from super_scad.type.Vector2 import Vector2
from super_scad.type.Vector3 import Vector3

ScadWidget = typing.NewType('ScadWidget', None)


class ScadWidget(ABC):
    """
    Abstract parent widget for all SuperSCAD widgets.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, args: Dict[str, Any] | None = None):
        """
        Object constructor.
        """
        self._args: Dict[str, Any] = {}
        """
        The arguments of this OpenSCAD widget.
        """

        self.__unit: Unit = Context.get_unit_length_current()
        """
        The unit of length of the Context of this OpenSCAD widget.
        """

        if args is not None:
            for key, value in args.items():
                if value is not None and value != self and key not in ('child', 'children'):
                    self._args[key] = value

        self._validate_arguments()

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        pass

    # ------------------------------------------------------------------------------------------------------------------
    @abstractmethod
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        raise NotImplementedError()

    # ------------------------------------------------------------------------------------------------------------------
    def uc(self, length: Any) -> Any:
        """
        Returns the length in the unit of the current context.

        :param length: The length.
        """
        if Context.get_unit_length_current() == self.__unit:
            if isinstance(length, int):
                return float(length)
            return length

        if length is None:
            return None

        if isinstance(length, float):
            return Length.convert(length, self.__unit, Context.get_unit_length_current())

        if isinstance(length, int):
            return Length.convert(float(length), self.__unit, Context.get_unit_length_current())

        if isinstance(length, Vector2):
            return Vector2(self.uc(length.x), self.uc(length.y))

        if isinstance(length, Vector3):
            return Vector3(self.uc(length.x), self.uc(length.y), self.uc(length.z))

        if isinstance(length, List):
            points = []
            for point in length:
                points.append(self.uc(point))

            return points

        raise ValueError('Can not convert length of type {}.'.format(type(length)))

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateRotate import PrivateRotate
from super_scad.type.Angle import Angle


class Rotate2D(ScadSingleChildParent):
    """
    Rotates its child about the z-axis. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#rotate.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param angle: The angle of rotation (around the z-axis).
        :param child: The child widget to be rotated (around the z-axis).
        :param child: The widget to be rotated.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle(self) -> float:
        """
        Returns the angle of rotation (around the z-axis).
        """
        return Angle.normalize(self._args.get('angle', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateRotate(angle=self.angle, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.boolean.Union import Union


class Compound(Union):
    """
    Creates a union of all its child widget. This is the sum of all children (logical or). See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/CSG_Modelling#union.
    """
    pass

# ----------------------------------------------------------------------------------------------------------------------



import inspect
import os
from pathlib import Path

from super_scad.scad.ScadCodeStore import ScadCodeStore
from super_scad.scad.Unit import Unit


class Context:
    """
    The context for generating OpenSCAD from SuperSCAD.
    """

    # ------------------------------------------------------------------------------------------------------------------
    __unit_length_current: Unit = Unit.FREE
    """
    The current unit of length.
    """

    __unit_length_final: Unit = Unit.FREE
    """
    The unit of length used in the generated OpenSCAD code.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 unit_length_final: Unit = Unit.MM,
                 fa: float = 12.0,
                 fs: float = 2.0,
                 fn: int = 0,
                 eps: float = 1E-2,
                 angle_digits: int = 4,
                 length_digits: int = 4,
                 scale_digits: int = 4):
        """
        Object constructor.

        :param unit_length_final: The unit of length used in the generated OpenSCAD code.
        :param fa: The minimum angle (in degrees) of each fragment. Known in OpenSCAD as $fa,
                   see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa.
        :param fs: The minimum circumferential length of each fragment. Known in OpenSCAD as $fs,
                   see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fs.
        :param fn: The number of fragments in 360 degrees. Values of 3 or more override $fa and $fs. Known in OpenSCAD
                   as $fn, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fn.
        :param eps: Epsilon value for clear overlap.
        :param angle_digits: The number of decimal places of an angle in the generated OpenSCAD code.
        :param length_digits: The number of decimal places of a length in the generated OpenSCAD code.
        :param scale_digits:  The number of decimal places of a scale or factor in the generated OpenSCAD code.
        """
        self.__project_home: Path = Path(os.getcwd()).resolve()
        """
        The home folder of the current project. 
        """

        self.__target_path: Path | None = None
        """
        The path to the OpenSCAD script that currently been generated.
        """

        self.__code_store: ScadCodeStore = ScadCodeStore()
        """
        The place were we store the generated OpenSCAD code.
        """

        self.__fa: float = fa
        """
        The minimum angle (in degrees) of each fragment. 
        Known in OpenSCAD as $fa, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa.
        """

        self.__fs: float = fs
        """
        The minimum circumferential length of each fragment.
        Known in OpenSCAD as $fs, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fs.
        """

        self.__fn: int = fn
        """
        The number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        Known in OpenSCAD as $fn, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fn.
        """

        self.__eps: float = eps
        """
        Epsilon value for clear overlap.
        """

        self.__unit_length_final: Unit = unit_length_final
        """
        The unit of length.
        """

        self.__angle_digits = angle_digits
        """
        The number of decimal places of an angle in the generated OpenSCAD code.
        """

        self.__length_digits = length_digits
        """
        The number of decimal places of a length in the generated OpenSCAD code.
        """

        self.__scale_digits = scale_digits
        """
        The number of decimal places of a scale or factor in the generated OpenSCAD code.
        """

        Context.set_unit_length_current(unit_length_final)
        Context.__set_unit_length_final(unit_length_final)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def project_home(self) -> Path:
        """
        Returns the current project's home directory.
        """
        return self.__project_home

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def target_path(self) -> Path | None:
        """
        Returns the path to the OpenSCAD script that currently been generated.
        """
        return self.__target_path

    # ------------------------------------------------------------------------------------------------------------------
    @target_path.setter
    def target_path(self, target_path: str) -> None:
        """
        Set the path to the OpenSCAD script that currently been generated.
        """
        self.__target_path = Path(os.path.realpath(target_path))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_digits(self) -> int:
        """
        Returns the number of decimal places of an angle in the generated OpenSCAD code.
        """
        return self.__angle_digits

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def length_digits(self) -> int:
        """
        Returns the number of decimal places of a length in the generated OpenSCAD code.
        """
        return self.__length_digits

    # ------------------------------------------------------------------------------------------------------------------
    def resolve_path(self, path: Path | str) -> Path:
        """
        Resolve a path relative from the caller script to a path relative to the project home.

        :param Path path: The path to resolve.
        """
        caller = Path(inspect.stack()[1].filename)
        absolute_path = Path(caller.parent.joinpath(path).resolve())

        if os.path.commonprefix([absolute_path, self.__project_home]) == str(self.__project_home):
            # works with python >=3.12 return absolute_path.relative_to(self.target_path.parent, walk_up=True)
            return Path(os.path.relpath(absolute_path, self.target_path.parent))

        return absolute_path

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def code_store(self) -> ScadCodeStore:
        """
        Returns code store.
        """
        return self.__code_store

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def eps(self) -> float:
        """
        Returns the epsilon value for clear overlap.
        """
        return self.__eps

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def resolution(self) -> float:
        """
        Returns the resolution of lengths in generated OpenSCAD code.
        """
        return 10 ** -self.__length_digits

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float:
        """
        Returns the minimum angle (in degrees) of each fragment.
        Known in OpenSCAD as $fa, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fa.
        """
        return self.__fa

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float:
        """
        Returns the minimum circumferential length of each fragment.
        Known in OpenSCAD as $fs, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fs.
        """
        return self.__fs

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int:
        """
        Returns the number of fragments in 360 degrees. Values of three or more override $fa and $fs.
        Known in OpenSCAD as $fn, see https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Other_Language_Features#$fn.
        """
        return self.__fn

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def __set_unit_length_final(__unit_length_final: Unit) -> None:
        """
        Sets the unit of length used in the generated OpenSCAD code.
        """
        Context.__unit_length_final = __unit_length_final

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_unit_length_final() -> Unit:
        """
        Returns the unit of length used in the generated OpenSCAD code.
        """
        return Context.__unit_length_final

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def set_unit_length_current(unit_length_current: Unit) -> None:
        """
        Sets the current unit of length.

        :param unit_length_current: The new current unit of length.
        """
        Context.__unit_length_current = unit_length_current

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def get_unit_length_current() -> Unit:
        """
        Returns the current unit of length.
        """
        return Context.__unit_length_current

    # ------------------------------------------------------------------------------------------------------------------
    def round_angle(self, angle: float) -> str:
        """
        Returns an angle rounded to the desired number of digits.

        :param angle: The length.
        """
        return str(round(float(angle), self.__angle_digits))

    # ------------------------------------------------------------------------------------------------------------------
    def round_length(self, length: float) -> str:
        """
        Returns a length rounded to the desired number of digits.

        :param length: The length.
        """
        return str(round(float(length), self.__length_digits))

    # ------------------------------------------------------------------------------------------------------------------
    def round_scale(self, scale: float) -> str:
        """
        Returns a scale or factor rounded to the desired number of digits.

        :param scale: The scale or factor.
        """
        return str(round(float(scale), self.__scale_digits))

# ----------------------------------------------------------------------------------------------------------------------



import random
from typing import List

from super_scad.d2.private.PrivatePolygon import PrivatePolygon
from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Angle import Angle
from super_scad.type.Vector2 import Vector2


class Polygon(ScadWidget):
    """
    Widget for creating polygons. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#polygon.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 primary: List[Vector2] | None = None,
                 points: List[Vector2] | None = None,
                 secondary: List[Vector2] | None = None,
                 secondaries: List[List[Vector2]] | None = None,
                 convexity: int | None = None):
        """
        Object constructor.

        :param primary: The list of 2D points of the polygon.
        :param points: Alias for primary.
        :param secondary: The secondary path that will be subtracted from the polygon.
        :param secondaries: The secondary paths that will be subtracted form the polygon.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line
                          through the child widget.
        """
        ScadWidget.__init__(self, args=locals())

        self._inner_angles: List[float] | None = None
        """
        The inner angels of the polygon (in the same order as the primary points).
        """

        self._normal_angles: List[float] | None = None
        """
        The absolute angel of the normal of each node.
        """

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'primary'}, {'points'})
        admission.validate_exclusive({'secondary'}, {'secondaries'})
        admission.validate_required({'primary', 'points'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def inner_angels(self) -> List[float]:
        """
        Returns the inner angels of the polygon (in the same order as the primary points).
        """
        if self._inner_angles is None:
            self._compute_angles()

        return self._inner_angles

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def normal_angels(self) -> List[float]:
        """
        Returns the absolute angel of the normal of each node.
        """
        if self._normal_angles is None:
            self._compute_angles()

        return self._normal_angles

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def primary(self) -> List[Vector2]:
        """
        Returns the points of the polygon.
        """
        return self.uc(self._args.get('primary', self._args.get('points')))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def secondaries(self) -> List[List[Vector2]] | None:
        """
        Returns the points of the polygon.
        """
        if 'secondaries' in self._args:
            tmp = []
            for points in self._args['secondaries']:
                tmp.append(self.uc(points))

            return tmp

        if 'secondary' in self._args:
            return [self.uc(self._args['secondary'])]

        return None

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the convexity of the polygon.
        """
        return self._args.get('convexity')

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        secondaries = self.secondaries
        if secondaries is None:
            return PrivatePolygon(points=self.primary, convexity=self.convexity)

        points = self.primary
        n = 0
        m = n + len(points)
        paths = [list(range(n, m))]
        n = m

        for secondary in secondaries:
            m = n + len(secondary)
            points += secondary
            paths.append(list(range(n, m)))
            n = m

        return PrivatePolygon(points=points, paths=paths, convexity=self.convexity)

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _count_intersections(nodes: List[Vector2], p1: Vector2, q1: Vector2) -> int:
        """
        Returns the number of intersections between a line segment (p1, q1) and the vertices of the polygon.

        @param nodes: The nodes of the polygon.
        @param p1: Start point of the line segment.
        @param q1: End point of the line segment.
        """
        intersections = 0

        n = len(nodes)
        for i in range(n):
            p2 = nodes[i]
            q2 = nodes[(i + 1) % n]

            if Vector2.do_intersect(p1, q1, p2, q2):
                intersections += 1

        return intersections

    # ------------------------------------------------------------------------------------------------------------------
    def _compute_angles(self) -> None:
        """
        Returns the inner angels of the polygon (in the same order as the primary points).
        """
        self._inner_angles = []
        self._normal_angles = []

        radius: float = 0.0
        nodes = self.primary
        for point in nodes:
            radius = max(radius, point.x, point.y)

        n = len(nodes)
        for i in range(n):
            p1 = nodes[(i - 1) % n]
            p2 = nodes[i]
            p3 = nodes[(i + 1) % n]

            q1 = Vector2((p1.x + p2.x + p3.x) / 3, (p1.y + p2.y + p3.y) / 3)
            q2 = Vector2.from_polar_coordinates(2.0 * radius, random.uniform(0.0, 360.0))

            number_of_intersections = Polygon._count_intersections(nodes, q1, q2)
            inner_angle = Vector2.angle_3p(p1, p2, p3)
            normal_angle = (p3 - p2).angle + 0.5 * inner_angle
            if number_of_intersections % 2 == 0:
                inner_angle = 360.0 - inner_angle
                normal_angle = 180.0 - normal_angle

            self._inner_angles.append(inner_angle)
            self._normal_angles.append(Angle.normalize(normal_angle))

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateMirror import PrivateMirror
from super_scad.type.Vector2 import Vector2


class Mirror2D(ScadSingleChildParent):
    """
    Transforms the child widget to a mirror of the original, as if it were the mirror image seen through a plane
    intersecting the origin. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#mirror.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 vector: Vector2 | None = None,
                 x: float | None = None,
                 y: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param vector:  The normal vector of the origin-intersecting mirror plane used, meaning the vector coming
                        perpendicularly out of the plane. Each coordinate of the original widget is altered such that
                        it becomes equidistant on the other side of this plane from the closest point on the plane.
        :param x:  The x-coordinate of the origin-intersecting mirror plane.
        :param y:  The y-coordinate of the origin-intersecting mirror plane.
        :param child: The widget to be mirrored.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

        self.__vector: Vector2 | None = None
        """
        The normalized normal vector of the origin-intersecting mirror plane used.
        """

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'vector'}, {'x', 'y'})
        admission.validate_required({'x', 'y', 'vector'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector2:
        """
        The normal vector of the origin-intersecting mirror plane.
        """
        if self.__vector is None:
            if 'vector' in self._args:
                self.__vector = self.uc(self._args['vector']).normal

            if 'x' in self._args or 'y' in self._args or 'z' in self._args:
                self.__vector = self.uc(Vector2(self._args.get('x', 0.0),
                                                self._args.get('y', 0.0))).normal

            self.__vector = self.__vector * (-1.0 if self.__vector.x < 0.0 else 1.0)

        return self.__vector

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateMirror(vector=self.vector, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import Dict, Set

from super_scad.private.PrivateSingleChildOpenScadCommand import PrivateSingleChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Vector2 import Vector2


class LinearExtrude(PrivateSingleChildOpenScadCommand):
    """
    Linear Extrusion is an operation that takes a 2D object as input and generates a 3D object as a result. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#linear_extrude.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 height: float,
                 center: bool = False,
                 convexity: int | None = None,
                 twist: float = 0.0,
                 scale: float | Vector2 = 1.0,
                 slices: int | None = None,
                 segments: int | None = None,
                 fa: float | None = None,
                 fs: float | None = None,
                 fn: int | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param height: The height of the extruded object.
        :param center: Whether the cylinder is centered along the z-as.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        :param twist: The number of degrees through which the shape is extruded. Setting the parameter twist = 360
                      extrudes through one revolution. The twist direction follows the left-hand rule.
        :param scale: Scales the 2D shape by this value over the height of the extrusion.
        :param slices: Defines the number of intermediate points along the Z axis of the extrusion. Its default
                       increases with the value of twist. Explicitly setting slices may improve the output refinement.
        :param segments: Adds vertices (points) to the extruded polygon resulting in smoother twisted geometries.
                         Segments need to be a multiple of the polygon's fragments to have an effect (6 or 9... for a
                         circle($fn=3), 8,12... for a square()).
        :param fa: The minimum angle (in degrees) of each fragment.
        :param fs: The minimum circumferential length of each fragment.
        :param fn: The fixed number of fragments in 360 degrees. Values of 3 or more override fa and fs.
        """
        PrivateSingleChildOpenScadCommand.__init__(self, command='linear_extrude', args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {'fa': '$fa', 'fs': '$fs', 'fn': '$fn'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return {'$fa', 'twist'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'height', '$fs'}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_scales(self) -> Set[str]:
        """
        Returns the set with arguments that are scales and factors.
        """
        return {'scale'}

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def height(self) -> float:
        """
        Returns the height of the extruded object.
        """
        return self.uc(self._args.get('height', 0.0))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def center(self) -> bool:
        """
        Returns whether the extruded object is centered along the z-as.
        """
        return self._args['center']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def convexity(self) -> int | None:
        """
        Returns the number of "inward" curves, i.e., expected number of path crossings of an arbitrary line through the
        child widget.
        """
        return self._args.get('convexity')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def twist(self) -> float:
        """
        Returns the number of degrees of through which the shape is extruded. Setting the parameter twist = 360
        extrudes through one revolution. The twist direction follows the left-hand rule.
        """
        return self._args.get('twist')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def scale(self) -> float | Vector2:
        """
        Returns the number of degrees of through which the shape is extruded. Setting the parameter twist = 360
        extrudes through one revolution. The twist direction follows the left-hand rule.
        """
        return self._args.get('scale')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def slices(self) -> int | None:
        """
        Returns the number of intermediate points along the Z axis of the extrusion. Its default
        increases with the value of twist. Explicitly setting slices may improve the output refinement.
        """
        return self._args.get('slices')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def segments(self) -> int | None:
        """
        Returns the Adds vertices (points) to the extruded polygon resulting in smoother twisted geometries.
        """
        return self._args.get('segments')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fa(self) -> float | None:
        """
        Returns the minimum angle (in degrees) of each fragment.
        """
        return self._args.get('fa')

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fs(self) -> float | None:
        """
        Returns the minimum circumferential length of each fragment.
        """
        return self.uc(self._args.get('fs'))

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def fn(self) -> int | None:
        """
        Returns the fixed number of fragments in 360 degrees. Values of 3 or more override $fa and $fs.
        """
        return self._args.get('fn')

# ----------------------------------------------------------------------------------------------------------------------



import re
from typing import Any, Dict, List, Set, Tuple

from super_scad.scad.Context import Context
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.type.Color import Color
from super_scad.type.Vector2 import Vector2
from super_scad.type.Vector3 import Vector3


class PrivateOpenScadCommand(ScadWidget):
    """
    Widget for creating OpenSCAD commands.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, command: str, args: Dict[str, Any]):
        """
        Object constructor.

        :param command: The name of the OpenSCAD command.
        :param args: The arguments of the OpenSCAD command.
        """
        ScadWidget.__init__(self, args=args)

        self._command: str = command
        """
        The name of the OpenSCAD command.
        """

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return self

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def command(self) -> str:
        """
        Returns the name of the OpenSCAD command.
        """
        return self._command

    # ------------------------------------------------------------------------------------------------------------------
    def generate_args(self, context: Context) -> str:
        """
        Returns the arguments of the OpenSCAD command.
        """
        argument_map = self._argument_map()
        argument_angles = self._argument_angles()
        argument_lengths = self._argument_lengths()
        argument_scales = self._argument_scales()

        args_as_str = '('
        first = True
        for key, value in self._args.items():
            if not first:
                args_as_str += ', '
            else:
                first = False

            real_name = argument_map.get(key, key)
            if real_name in argument_angles:
                real_value = self.__format_argument(context, value, True, False, False)
            elif real_name in argument_lengths:
                real_value = self.__format_argument(context, self.uc(value), False, True, False)
            elif real_name in argument_scales:
                real_value = self.__format_argument(context, value, False, False, True)
            else:
                real_value = self.__format_argument(context, value, False, False, False)

            if real_name is None:
                args_as_str += '{}'.format(real_value)
            else:
                args_as_str += '{} = {}'.format(real_name, real_value)
        args_as_str += ')'

        return args_as_str

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_map(self) -> Dict[str, str | None]:
        """
        Returns the map from SuperSCAD arguments to OpenSCAD arguments.
        """
        return {}

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_angles(self) -> Set[str]:
        """
        Returns the set with arguments that are angles.
        """
        return set()

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return set()

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_scales(self) -> Set[str]:
        """
        Returns the set with arguments that are scales and factors.
        """
        return set()

    # ------------------------------------------------------------------------------------------------------------------
    def __format_argument(self,
                          context: Context,
                          argument: Any,
                          is_angle: bool,
                          is_length: bool,
                          is_scale: bool) -> str:
        """
        Returns an argument of the OpenSCAD command.

        :param context: The build context.
        :param argument: The argument of OpenSCAD command.
        """
        if isinstance(argument, float):
            if is_length:
                argument = context.round_length(argument)
            elif is_angle:
                argument = context.round_angle(argument)
            elif is_scale:
                argument = context.round_scale(argument)
            if argument == '-0.0':
                argument = '0.0'

            return argument

        if isinstance(argument, Vector2):
            return "[{}, {}]".format(self.__format_argument(context, float(argument.x), is_angle, is_length, is_scale),
                                     self.__format_argument(context, float(argument.y), is_angle, is_length, is_scale))

        if isinstance(argument, Vector3):
            return "[{}, {}, {}]".format(
                    self.__format_argument(context, float(argument.x), is_angle, is_length, is_scale),
                    self.__format_argument(context, float(argument.y), is_angle, is_length, is_scale),
                    self.__format_argument(context, float(argument.z), is_angle, is_length, is_scale))

        if isinstance(argument, bool):
            return str(argument).lower()

        if isinstance(argument, str):
            return '"{}"'.format(re.sub(r'([\\\"])', r'\\\1', argument))

        if isinstance(argument, int):
            return str(argument)

        if isinstance(argument, List):
            parts = []
            for element in argument:
                parts.append(self.__format_argument(context, element, is_angle, is_length, is_scale))

            return '[{}]'.format(', '.join(parts))

        if isinstance(argument, Tuple):
            parts = []
            for element in argument:
                parts.append(self.__format_argument(context, element, is_angle, is_length, is_scale))

            return '[{}]'.format(', '.join(parts))

        if isinstance(argument, Color):
            return str(argument)

        raise ValueError(f'Can not format argument of type {argument.__class__}.')

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.private.PrivateMultiChildOpenScadCommand import PrivateMultiChildOpenScadCommand
from super_scad.scad.ScadWidget import ScadWidget


class Intersection(PrivateMultiChildOpenScadCommand):
    """
    Creates the intersection of all child widgets. This keeps the overlapping portion (logical and). See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/CSG_Modelling#intersection.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, *, children: List[ScadWidget]):
        """
        Object constructor.

        :param children: The child widgets.
        """
        PrivateMultiChildOpenScadCommand.__init__(self, command='intersection', args=locals(), children=children)

# ----------------------------------------------------------------------------------------------------------------------



from typing import List

from super_scad.scad.Unit import Unit
from super_scad.scad.unit.AngstromUnit import AngstromUnit
from super_scad.scad.unit.AstronomicalUnit import AstronomicalUnit
from super_scad.scad.unit.AttoparsecUnit import AttoparsecUnit
from super_scad.scad.unit.CentimeterUnit import CentimeterUnit
from super_scad.scad.unit.DecimeterUnit import DecimeterUnit
from super_scad.scad.unit.FootUnit import FootUnit
from super_scad.scad.unit.InchUnit import InchUnit
from super_scad.scad.unit.KilometerUnit import KilometerUnit
from super_scad.scad.unit.LightYearUnit import LightYearUnit
from super_scad.scad.unit.LiUnit import LiUnit
from super_scad.scad.unit.MeterUnit import MeterUnit
from super_scad.scad.unit.MicrometerUnit import MicrometerUnit
from super_scad.scad.unit.MileUnit import MileUnit
from super_scad.scad.unit.MillimeterUnit import MillimeterUnit
from super_scad.scad.unit.ParsecUnit import ParsecUnit
from super_scad.scad.unit.RoyalCubitUnit import RoyalCubitUnit
from super_scad.scad.unit.ThouUnit import ThouUnit
from super_scad.scad.unit.YardUnit import YardUnit


class Length:
    """
    Utility class for converting lengths between different units of length.
    """
    # ------------------------------------------------------------------------------------------------------------------
    __ratio: List[List[float | None]] = []
    """
    The ratios of all unit of lengths.
    """

    __units: List = [AngstromUnit,
                     AstronomicalUnit,
                     AttoparsecUnit,
                     CentimeterUnit,
                     DecimeterUnit,
                     FootUnit,
                     InchUnit,
                     KilometerUnit,
                     LightYearUnit,
                     LiUnit,
                     MeterUnit,
                     MicrometerUnit,
                     MileUnit,
                     MillimeterUnit,
                     ParsecUnit,
                     RoyalCubitUnit,
                     ThouUnit,
                     YardUnit]
    """
    All unit of unit of lengths known to SuperSCAD.
    """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def init_ratios() -> None:
        """
        Initializes the ratios of lengths.
        """
        max_value = max(unit.value for unit in Unit)
        for i in range(max_value + 1):
            for j in range(max_value + 1):
                if j == 0:
                    Length.__ratio.append([])
                Length.__ratio[i].append(None)

        for unit_from_class in Length.__units:
            unit_from = unit_from_class()
            id_from = unit_from.id()
            for unit_to_class in Length.__units:
                unit_to = unit_to_class()
                id_to = unit_to.id()
                Length.__ratio[id_from][id_to] = unit_from.meters() / unit_to.meters()

        # One special case. A length that is free of any scale can be converted to a length that is free of any scale
        # only.
        Length.__ratio[Unit.FREE.value][Unit.FREE.value] = 1.0

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def convert(length: float, from_unit: Unit, to_unit: Unit) -> float:
        """
        Converts a length from one unit to another.

        @param length: The length to convert.
        @param from_unit: The unit to convert from.
        @param to_unit: The unit to convert to.
        """
        try:
            return length * Length.__ratio[from_unit.value][to_unit.value]
        except Exception:
            raise ValueError(f'Cannot convert length {length} from {from_unit.name} to {to_unit.name}.')

# ----------------------------------------------------------------------------------------------------------------------



from typing import Any, List, Set


class ArgumentAdmission:
    """
    Class for validation arguments of a SupeSCAD widget constructor.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self, arguments: dict[str, Any]):
        """
        Object constructor.

        :param arguments: The arguments passed to the constructor of the SuperSCAD widget.
        """
        self.__arguments_set: Set[str] = set(arguments.keys())
        """
        A set with all the none empty arguments passed to the constructor of a ScadWidget.
        """

    # ------------------------------------------------------------------------------------------------------------------
    @staticmethod
    def _to_string(sets: List[Set[str]]) -> str:
        """

        """
        if len(sets) == 0:
            return ''

        if len(sets) == 1:
            return str(sets[0])

        if len(sets) == 2:
            return str(sets[0]) + ' and ' + str(sets[1])

        ret = ''
        for i in range(len(sets) - 1):
            ret += str(sets[i]) + ', '
        ret += ', and ' + str(sets[-1])

        return ret

    # ------------------------------------------------------------------------------------------------------------------
    def validate_exclusive(self, *args: Set[str]) -> None:
        """
        Validates that only arguments belonging to one of the given sets are passed to the constructor of the SuperScad
        widget.
        """
        supplied_sets = []
        for arg in args:
            supplied = set.intersection(self.__arguments_set, arg)
            if supplied:
                supplied_sets.append(supplied)

        if len(supplied_sets) > 1:
            sets = self._to_string(supplied_sets)
            raise ValueError('The following set of arguments are not exclusive: {}'.format(sets))

    # ------------------------------------------------------------------------------------------------------------------
    def validate_required(self, *args: Set[str]) -> None:
        """
        Validates that at least one argument belonging to each of the given sets are passed to the constructor of the
        SuperSCAD widget.
        """
        for arg in args:
            supplied = set.intersection(self.__arguments_set, arg)
            if not supplied:
                raise ValueError('At least one of these arguments {} must be supplied'.format(str(arg)))

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.d2.private.PrivateProjection import PrivateProjection
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.Translate3D import Translate3D


class Projection(ScadSingleChildParent):
    """
    Creates 2D drawings from 3D models. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Using_the_2D_Subsystem#3D_to_2D_Projection.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 cut: bool = False,
                 z: float | None = None,
                 child: ScadWidget):
        """
        Object constructor.

        :param cut: Whether to cut the 3D model.
        :param z: The height where the 3D model will be cut. Defaults to 0.0.
        :param child: The widget with the 3D model.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def cut(self) -> bool:
        """
        Returns whether to cut the 3D model.
        """
        return self._args['cut']

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def z(self) -> float | None:
        """
        Returns the height where the 3D model will be cut.
        """
        if not self.cut:
            return None

        return self._args.get('z', 0.0)

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        if self.z is None:
            child = self.child
        else:
            child = Translate3D(z=-self.z, child=self.child)

        return PrivateProjection(cut=self.cut, child=child)

# ----------------------------------------------------------------------------------------------------------------------



from typing import List, Set

from super_scad.private.PrivateOpenScadCommand import PrivateOpenScadCommand
from super_scad.type.Vector3 import Vector3


class PrivatePolyhedron(PrivateOpenScadCommand):
    """
    Widget for creating polyhedrons. See https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Primitive_Solids#polyhedron.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 points: List[Vector3],
                 faces: List[List[int]],
                 convexity: int | None = None):
        """
        Object constructor.

        :param points: String representation of a list of 3D points.
        :param faces:  String representation of the faces that collectively enclose the solid.
        :param convexity: Number of "inward" curves, i.e., expected number of path crossings of an arbitrary line 
                          through the child widget.
        """
        PrivateOpenScadCommand.__init__(self, command='polyhedron', args=locals())

    # ------------------------------------------------------------------------------------------------------------------
    def _argument_lengths(self) -> Set[str]:
        """
        Returns the set with arguments that are lengths.
        """
        return {'points'}

# ----------------------------------------------------------------------------------------------------------------------



from super_scad.scad.ArgumentAdmission import ArgumentAdmission
from super_scad.scad.Context import Context
from super_scad.scad.ScadSingleChildParent import ScadSingleChildParent
from super_scad.scad.ScadWidget import ScadWidget
from super_scad.transformation.private.PrivateRotate import PrivateRotate
from super_scad.type.Vector3 import Vector3


class Rotate3D(ScadSingleChildParent):
    """
    Rotates its child degrees about the axis of the coordinate system or around an arbitrary axis. See
    https://en.wikibooks.org/wiki/OpenSCAD_User_Manual/Transformations#rotate.
    """

    # ------------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 *,
                 angle: float | Vector3 | None = None,
                 angle_x: float | None = None,
                 angle_y: float | None = None,
                 angle_z: float | None = None,
                 vector: Vector3 | None = None,
                 child: ScadWidget) -> None:
        """
        Object constructor.

        :param angle: The angle of rotation around all axis or a vector.
        :param angle_x: The angle of rotation around the x-axis.
        :param angle_y: The angle of rotation around the y-axis.
        :param angle_z: The angle of rotation around the z-axis.
        :param vector: The vector of rotation.
        :param child: The widget to be rotated.
        """
        ScadSingleChildParent.__init__(self, args=locals(), child=child)

    # ------------------------------------------------------------------------------------------------------------------
    def _validate_arguments(self) -> None:
        """
        Validates the arguments supplied to the constructor of this SuperSCAD widget.
        """
        admission = ArgumentAdmission(self._args)
        admission.validate_exclusive({'angle', 'vector'}, {'angle_x', 'angle_y', 'angle_z'})
        admission.validate_required({'angle_x', 'angle_y', 'angle_z', 'angle'})

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle(self) -> float | Vector3 | None:
        """
        Returns angle of rotation around all axis or a vector.
        """
        if 'vector' in self._args:
            return self._args.get('angle')

        return Vector3(self.angle_x, self.angle_y, self.angle_z)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_x(self) -> float | None:
        """
        Returns the angle of rotation around the x-axis.
        """
        if 'vector' in self._args:
            return None

        if 'angle' in self._args:
            return self._args['angle'].x

        return self._args.get('angle_x', 0.0)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_y(self) -> float | None:
        """
        Returns the angle of rotation around the y-axis.
        """
        if 'vector' in self._args:
            return None

        if 'angle' in self._args:
            return self._args['angle'].y

        return self._args.get('angle_y', 0.0)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def angle_z(self) -> float | None:
        """
        Returns the angle of rotation around the z-axis.
        """
        if 'vector' in self._args:
            return None

        if 'angle' in self._args:
            return self._args['angle'].z

        return self._args.get('angle_z', 0.0)

    # ------------------------------------------------------------------------------------------------------------------
    @property
    def vector(self) -> Vector3 | None:
        """
        Returns the vector of rotation.
        """
        if 'vector' in self._args:
            return self.uc(self._args['vector'])

        return None

    # ------------------------------------------------------------------------------------------------------------------
    def build(self, context: Context) -> ScadWidget:
        """
        Builds a SuperSCAD widget.

        :param context: The build context.
        """
        return PrivateRotate(angle=self.angle, vector=self.vector, child=self.child)

# ----------------------------------------------------------------------------------------------------------------------
